"""Fake MRP Apple TV for tests."""

import asyncio
import logging
import struct
from datetime import datetime

from pyatv import const
from pyatv.support import log_protobuf
from pyatv.mrp import chacha20, messages, protobuf, variant
from pyatv.mrp.protobuf import CommandInfo_pb2 as cmd
from pyatv.mrp.protobuf import SetStateMessage as ssm
from pyatv.mrp.protobuf import SendCommandResultMessage as scr
from pyatv.mrp.server_auth import MrpServerAuth

_LOGGER = logging.getLogger(__name__)

_KEY_LOOKUP = {
    # name: [usage_page, usage, button hold time (seconds)]
    "up": [1, 0x8C, 0],
    "down": [1, 0x8D, 0],
    "left": [1, 0x8B, 0],
    "right": [1, 0x8A, 0],
    "stop": [12, 0xB7, 0],
    "next": [12, 0xB5, 0],
    "previous": [12, 0xB6, 0],
    "select": [1, 0x89, 0],
    "menu": [1, 0x86, 0],
    "topmenu": [12, 0x60, 0],
    "home": [12, 0x40, 1],
    "suspend": [1, 0x82, 0],
    "wakeup": [1, 0x83, 0],
    "volumeup": [12, 0xE9, 0],
    "volumedown": [12, 0xEA, 0],
}

_COMMAND_LOOKUP = {
    cmd.Play: "play",
    cmd.TogglePlayPause: "playpause",
    cmd.Pause: "pause",
    cmd.Stop: "stop",
    cmd.NextTrack: "nextitem",
    cmd.PreviousTrack: "previtem",
}

_REPEAT_LOOKUP = {
    const.RepeatState.Track: protobuf.RepeatMode.One,
    const.RepeatState.All: protobuf.RepeatMode.All,
}

_SHUFFLE_LOOKUP = {
    const.ShuffleState.Off: protobuf.ShuffleMode.Off,
    const.ShuffleState.Albums: protobuf.ShuffleMode.Albums,
    const.ShuffleState.Songs: protobuf.ShuffleMode.Songs,
}

_COCOA_BASE = (datetime(1970, 1, 1) - datetime(2001, 1, 1)).total_seconds()


DEVICE_NAME = "Fake MRP ATV"
PLAYER_IDENTIFIER = "com.github.postlund.pyatv"


def _convert_key_press(use_page, usage):
    for name, codes in _KEY_LOOKUP.items():
        if codes[0] == use_page and codes[1] == usage:
            return name
    raise Exception("unsupported key: use_page={0}, usage={1}".format(use_page, usage))


def _fill_item(item, metadata):
    if metadata.identifier:
        item.identifier = metadata.identifier

    md = item.metadata
    md.elapsedTimeTimestamp = _COCOA_BASE
    if metadata.artist:
        md.trackArtistName = metadata.artist
    if metadata.album:
        md.albumName = metadata.album
    if metadata.title:
        md.title = metadata.title
    if metadata.genre:
        md.genre = metadata.genre
    if metadata.total_time is not None:
        md.duration = metadata.total_time
    if metadata.position is not None:
        md.elapsedTime = metadata.position
    if metadata.playback_rate is not None:
        md.playbackRate = metadata.playback_rate
    if metadata.media_type:
        md.mediaType = metadata.media_type
    if metadata.artwork_mimetype:
        md.artworkAvailable = True
        md.artworkMIMEType = metadata.artwork_mimetype
        if metadata.artwork_identifier:
            md.artworkIdentifier = metadata.artwork_identifier


def _set_state_message(metadata, identifier):
    # Most things are hardcoded here for simplicity. Will change that
    # as time goes by and more dynamic content is needed.
    set_state = messages.create(protobuf.SET_STATE_MESSAGE)
    inner = set_state.inner()
    inner.playbackState = metadata.playback_state
    inner.displayName = "Fake Player"

    if metadata.supported_commands:
        for command in metadata.supported_commands:
            item = inner.supportedCommands.supportedCommands.add()
            item.command = command
            item.enabled = True

            if command in [
                protobuf.CommandInfo_pb2.SkipForward,
                protobuf.CommandInfo_pb2.SkipBackward,
            ]:
                if metadata.skip_time is not None:
                    item.preferredIntervals.append(metadata.skip_time)

    if metadata.repeat and metadata.repeat != const.RepeatState.Off:
        cmd = inner.supportedCommands.supportedCommands.add()
        cmd.command = protobuf.CommandInfo_pb2.ChangeRepeatMode
        cmd.repeatMode = _REPEAT_LOOKUP[metadata.repeat]

    if metadata.shuffle:
        cmd = inner.supportedCommands.supportedCommands.add()
        cmd.command = protobuf.CommandInfo_pb2.ChangeShuffleMode
        cmd.shuffleMode = _SHUFFLE_LOOKUP.get(metadata.shuffle)

    queue = inner.playbackQueue
    queue.location = 0
    _fill_item(queue.contentItems.add(), metadata)

    client = inner.playerPath.client
    client.processIdentifier = 123
    client.bundleIdentifier = identifier
    return set_state


class PlayingState:
    def __init__(self, **kwargs):
        """Initialize a new PlayingState."""
        self.identifier = kwargs.get("identifier")
        self.playback_state = kwargs.get("playback_state")
        self.title = kwargs.get("title")
        self.artist = kwargs.get("artist")
        self.album = kwargs.get("album")
        self.genre = kwargs.get("genre")
        self.total_time = kwargs.get("total_time")
        self.position = kwargs.get("position")
        self.repeat = kwargs.get("repeat")
        self.shuffle = kwargs.get("shuffle")
        self.media_type = kwargs.get("media_type")
        self.playback_rate = kwargs.get("playback_rate")
        self.supported_commands = kwargs.get("supported_commands")
        self.artwork = kwargs.get("artwork")
        self.artwork_mimetype = kwargs.get("artwork_mimetype")
        self.artwork_width = kwargs.get("artwork_width")
        self.artwork_height = kwargs.get("artwork_height")
        self.skip_time = kwargs.get("skip_time")


class FakeMrpState:
    def __init__(self):
        """State of a fake MRP device."""
        self.clients = []
        self.outstanding_keypresses = set()  # Pressed but not released
        self.last_button_pressed = None
        self.connection_state = None
        self.states = {}
        self.active_player = None
        self.powered_on = True
        self.has_authenticated = False

    def _send(self, msg):
        for client in self.clients:
            client.send(msg)

    # This is a hack for now (if anyone has paired, OK...)
    @property
    def has_paired(self):
        return any([client.has_paired for client in self.clients])

    def update_state(self, identifier):
        state = self.states[identifier]
        self._send(_set_state_message(state, identifier))

    def set_player_state(self, identifier, state):
        self.states[identifier] = state
        self.update_state(identifier)

    def get_player_state(self, identifier):
        return self.states.get(identifier)

    def set_active_player(self, identifier):
        if identifier is not None and identifier not in self.states:
            raise Exception("invalid player: %s", identifier)

        self.active_player = identifier
        now_playing = messages.create(protobuf.SET_NOW_PLAYING_CLIENT_MESSAGE)
        client = now_playing.inner().client
        if identifier:
            client.bundleIdentifier = identifier
        self._send(now_playing)

    def item_update(self, metadata, identifier):
        msg = messages.create(protobuf.UPDATE_CONTENT_ITEM_MESSAGE)
        inner = msg.inner()

        _fill_item(inner.contentItems.add(), metadata)

        client = inner.playerPath.client
        client.processIdentifier = 123
        client.bundleIdentifier = identifier

        state = self.get_player_state(identifier)
        for var, value in vars(metadata).items():
            if value:
                setattr(state, var, value)

        self._send(msg)

    def update_client(self, display_name, identifier):
        msg = messages.create(protobuf.UPDATE_CLIENT_MESSAGE)
        client = msg.inner().client
        client.bundleIdentifier = identifier
        client.displayName = display_name
        self._send(msg)

    def volume_control(self, available):
        msg = messages.create(protobuf.VOLUME_CONTROL_AVAILABILITY_MESSAGE)
        msg.inner().volumeControlAvailable = available
        self._send(msg)


class FakeMrpServiceFactory:
    def __init__(self, state, app, loop):
        self.state = state
        self.app = app
        self.loop = loop
        self.server = None

    async def start(self, start_web_server: bool):
        coro = self.loop.create_server(
            lambda: FakeMrpService(self.state, self.app, self.loop), "0.0.0.0"
        )
        self.server = await self.loop.create_task(coro)
        _LOGGER.info("Started MRP server at port %d", self.port)

    async def cleanup(self):
        if self.server:
            self.server.close()

    @property
    def port(self):
        return self.server.sockets[0].getsockname()[1]


class FakeMrpService(MrpServerAuth, asyncio.Protocol):
    """Implementation of a fake MRP Apple TV."""

    def __init__(self, state, app, loop):
        MrpServerAuth.__init__(self, self, DEVICE_NAME)
        self.state = state
        self.state.clients.append(self)
        self.app = app
        self.loop = loop

        self.buffer = b""
        self.chacha = None
        self.transport = None

    def connection_made(self, transport):
        _LOGGER.debug("Client connected")
        self.transport = transport

    def connection_lost(self, exc):
        _LOGGER.debug("Client disconnected")
        self.transport = None
        self.state.clients.remove(self)

    def enable_encryption(self, input_key, output_key):
        """Enable encryption with specified keys."""
        self.chacha = chacha20.Chacha20Cipher(input_key, output_key)
        self.state.has_authenticated = True

    def send(self, message):
        if not self.transport:
            return

        data = message.SerializeToString()
        if self.chacha:
            data = self.chacha.encrypt(data)

        length = variant.write_variant(len(data))
        self.transport.write(length + data)
        log_protobuf(_LOGGER, ">> Send: Protobuf", message)

    def _send_device_info(self, identifier=None, update=False):
        resp = messages.device_information(DEVICE_NAME, "1234", update=update)
        if identifier:
            resp.identifier = identifier
        resp.inner().logicalDeviceCount = 1 if self.state.powered_on else 0
        self.send(resp)

    def data_received(self, data):
        self.buffer += data

        while self.buffer:
            length, raw = variant.read_variant(self.buffer)
            if len(raw) < length:
                return

            data = raw[:length]
            self.buffer = raw[length:]
            if self.chacha:
                data = self.chacha.decrypt(data)

            parsed = protobuf.ProtocolMessage()
            parsed.ParseFromString(data)
            log_protobuf(_LOGGER, "<< Receive: Protobuf", parsed)

            try:
                name = parsed.Type.Name(parsed.type).lower().replace("_message", "")

                _LOGGER.debug("Received %s message", name)
                getattr(self, "handle_" + name)(parsed, parsed.inner())
            except AttributeError:
                _LOGGER.exception("No message handler for " + str(parsed))
            except Exception:
                _LOGGER.exception("Error while dispatching message")

    def handle_device_info(self, message, inner):
        self._send_device_info(update=False, identifier=message.identifier)

    def handle_set_connection_state(self, message, inner):
        _LOGGER.debug("Changed connection state to %d", inner.state)
        self.state.connection_state = inner.state

    def handle_client_updates_config(self, message, inner):
        for identifier, metadata in self.state.states.items():
            self.send(_set_state_message(metadata, identifier))

        # Trigger sending of SetNowPlayingClientMessage
        if self.state.active_player:
            self.state.set_active_player(self.state.active_player)

    def handle_get_keyboard_session(self, message, inner):
        # This message has a lot more fields, but pyatv currently
        # not use them so ignore for now
        self.send(
            messages.create(protobuf.KEYBOARD_MESSAGE, identifier=message.identifier)
        )

    def handle_send_hid_event(self, message, inner):
        # These corresponds to the bytes mapping to pressed key (see
        # send_hid_event in pyatv/mrp/messages.py)
        start = inner.hidEventData[43:49]
        use_page, usage, down_press = struct.unpack(">HHH", start)

        if down_press == 1:
            self.state.outstanding_keypresses.add((use_page, usage))
            self.send(messages.create(0, identifier=message.identifier))
        elif down_press == 0:
            if (use_page, usage) in self.state.outstanding_keypresses:
                if (
                    _convert_key_press(use_page, usage) == "select"
                    and self.state.last_button_pressed == "home"
                ):
                    self.state.powered_on = False
                    self._send_device_info(update=True)
                self.state.last_button_pressed = _convert_key_press(use_page, usage)
                self.state.outstanding_keypresses.remove((use_page, usage))
                _LOGGER.debug("Pressed button: %s", self.state.last_button_pressed)
                self.send(messages.create(0, identifier=message.identifier))
            else:
                _LOGGER.error("Missing key down for %d,%d", use_page, usage)
        else:
            _LOGGER.error("Invalid key press state: %d", down_press)

    def handle_send_command(self, message, inner):
        state = self.state.get_player_state(self.state.active_player)
        button = _COMMAND_LOOKUP.get(inner.command)
        if button:
            self.state.last_button_pressed = button
            _LOGGER.debug("Pressed button: %s", self.state.last_button_pressed)
        elif inner.command == cmd.ChangeRepeatMode:
            state.repeat = {
                protobuf.RepeatMode.One: const.RepeatState.Track,
                protobuf.RepeatMode.All: const.RepeatState.All,
            }.get(inner.options.repeatMode, const.RepeatState.Off)
            self.state.update_state(self.state.active_player)
            _LOGGER.debug("Change repeat state to %s", state.repeat)
        elif inner.command == cmd.ChangeShuffleMode:
            state.shuffle = {
                protobuf.ShuffleMode.Off: const.ShuffleState.Off,
                protobuf.ShuffleMode.Albums: const.ShuffleState.Albums,
                protobuf.ShuffleMode.Songs: const.ShuffleState.Songs,
            }.get(inner.options.shuffleMode, const.ShuffleState.Off)
            self.state.update_state(self.state.active_player)
            _LOGGER.debug("Change shuffle state to %s", state.shuffle)
        elif inner.command == cmd.SeekToPlaybackPosition:
            state.position = inner.options.playbackPosition
            self.state.update_state(self.state.active_player)
            _LOGGER.debug("Seek to position: %d", state.position)
        elif inner.command == cmd.SkipForward:
            state.position += int(inner.options.skipInterval)
            self.state.update_state(self.state.active_player)
            _LOGGER.debug("Skip forward %ds", inner.options.skipInterval)
        elif inner.command == cmd.SkipBackward:
            state.position -= int(inner.options.skipInterval)
            self.state.update_state(self.state.active_player)
            _LOGGER.debug("Skip backwards %d", inner.options.skipInterval)
        else:
            _LOGGER.warning("Unhandled button press: %s", message.inner().command)
            self.send(
                messages.command_result(
                    message.identifier, send_error=protobuf.SendError.NoCommandHandlers
                )
            )
            return

        self.send(messages.command_result(message.identifier))

    def handle_playback_queue_request(self, message, inner):
        state = self.state.get_player_state(self.state.active_player)

        setstate = messages.create(
            protobuf.SET_STATE_MESSAGE, identifier=message.identifier
        )
        queue = setstate.inner().playbackQueue
        queue.location = 0
        item = queue.contentItems.add()
        item.artworkData = self.state.states[self.state.active_player].artwork
        item.artworkDataWidth = state.artwork_width or 456
        item.artworkDataHeight = state.artwork_height or 789
        self.send(setstate)

    def handle_wake_device(self, message, inner):
        self.send(messages.command_result(message.identifier))
        self.state.powered_on = True
        self._send_device_info(update=True)


class FakeMrpUseCases:
    """Wrapper for altering behavior of a FakeMrpAppleTV instance."""

    def __init__(self, state):
        """Initialize a new FakeMrpUseCases."""
        self.state = state

    def change_volume_control(self, available):
        """Change volume control availability."""
        self.state.volume_control(available)

    def change_artwork(
        self, artwork, mimetype, identifier="artwork", width=None, height=None
    ):
        """Call to change artwork response."""
        metadata = self.state.get_player_state(PLAYER_IDENTIFIER)
        metadata.artwork = artwork
        metadata.artwork_mimetype = mimetype
        metadata.artwork_identifier = identifier
        metadata.artwork_width = width
        metadata.artwork_height = height
        self.state.update_state(PLAYER_IDENTIFIER)

    def change_metadata(self, **kwargs):
        """Change metadata for item via ContentItemUpdate."""
        metadata = self.state.get_player_state(PLAYER_IDENTIFIER)

        # Update saved metadata
        for key, value in kwargs.items():
            setattr(metadata, key, value)

        # Create a temporary metadata instance with requested parameters
        change = PlayingState(**kwargs)
        self.state.item_update(change, PLAYER_IDENTIFIER)

    def change_state(self, **kwargs):
        """Update playing state and set SetStateMessage."""
        metadata = self.state.get_player_state(PLAYER_IDENTIFIER)

        # Update saved metadata
        for key, value in kwargs.items():
            setattr(metadata, key, value)

        self.state.update_state(PLAYER_IDENTIFIER)

    def update_client(self, display_name):
        """Update playing client with new information."""
        self.state.update_client(display_name, PLAYER_IDENTIFIER)

    def nothing_playing(self):
        """Call to put device in idle state."""
        self.state.set_active_player(None)

    def example_video(self, **kwargs):
        """Play some example video."""
        kwargs.setdefault("title", "dummy")
        kwargs.setdefault("paused", True)
        kwargs.setdefault("position", 3)
        kwargs.setdefault("total_time", 123)
        self.video_playing(**kwargs)

    def video_playing(self, paused, title, total_time, position, **kwargs):
        """Call to change what is currently plaing to video."""
        metadata = PlayingState(
            playback_state=ssm.Paused if paused else ssm.Playing,
            title=title,
            total_time=total_time,
            position=position,
            media_type=protobuf.ContentItemMetadata.Video,
            **kwargs,
        )
        self.state.set_player_state(PLAYER_IDENTIFIER, metadata)
        self.state.set_active_player(PLAYER_IDENTIFIER)

    def example_music(self, **kwargs):
        """Play some example music."""
        kwargs.setdefault("paused", True)
        kwargs.setdefault("title", "music")
        kwargs.setdefault("artist", "artist")
        kwargs.setdefault("album", "album")
        kwargs.setdefault("total_time", 49)
        kwargs.setdefault("position", 22)
        kwargs.setdefault("genre", "genre")
        self.music_playing(**kwargs)

    def music_playing(
        self, paused, artist, album, title, genre, total_time, position, **kwargs
    ):
        """Call to change what is currently plaing to music."""
        metadata = PlayingState(
            playback_state=ssm.Paused if paused else ssm.Playing,
            artist=artist,
            album=album,
            title=title,
            genre=genre,
            total_time=total_time,
            position=position,
            media_type=protobuf.ContentItemMetadata.Music,
            **kwargs,
        )
        self.state.set_player_state(PLAYER_IDENTIFIER, metadata)
        self.state.set_active_player(PLAYER_IDENTIFIER)

    def media_is_loading(self):
        """Call to put device in a loading state."""
        metadata = PlayingState(playback_state=ssm.Interrupted)
        self.state.set_player_state(PLAYER_IDENTIFIER, metadata)
        self.state.set_active_player(PLAYER_IDENTIFIER)



"""Unit tests for pyatv.interface."""

import pytest
from typing import Dict

from pyatv import convert, exceptions, interface
from pyatv.interface import FeatureInfo
from pyatv.const import (
    MediaType,
    DeviceState,
    RepeatState,
    ShuffleState,
    OperatingSystem,
    DeviceModel,
    FeatureState,
    FeatureName,
)


class TestClass:

    variable = 1234

    def test_method(self):
        """Help text."""
        pass

    def another_method(self):
        """Some other help text. More text here. Test."""
        pass

    @property
    def some_property(self):
        """Property help"""
        pass

    def abbrev_help(self):
        """Type, e.g. a, b or c."""
        pass

    def abbrev_help_more_text(self):
        """Type, e.g. a, b or c. Some other text."""
        pass

    def _private_method_ignored(self):
        """Not parsed."""
        pass


class PlayingDummy(interface.Playing):
    def __init__(
        self,
        media_type=MediaType.Video,
        device_state=DeviceState.Playing,
        title=None,
        artist=None,
        album=None,
        genre=None,
        total_time=None,
        position=None,
        shuffle=ShuffleState.Songs,
        repeat=RepeatState.Track,
    ):
        self._media_type = media_type
        self._device_state = device_state
        self._title = title
        self._artist = artist
        self._album = album
        self._genre = genre
        self._total_time = total_time
        self._position = position
        self._shuffle = shuffle
        self._repeat = repeat

    @property
    def media_type(self):
        """What type of media is currently playing, e.g. video, music."""
        return self._media_type

    @property
    def device_state(self):
        """Current device state, e.g. playing or paused."""
        return self._device_state

    @property
    def title(self):
        """Title of the current media, e.g. movie or song name."""
        return self._title

    @property
    def artist(self):
        """Artist of the currently playing song."""
        return self._artist

    @property
    def album(self):
        """Album of the currently playing song."""
        return self._album

    @property
    def genre(self):
        """Genre of the currently playing song."""
        return self._genre

    @property
    def total_time(self):
        """Total play time in seconds."""
        return self._total_time

    @property
    def position(self):
        """Current position in the playing media (seconds)."""
        return self._position

    @property
    def shuffle(self):
        """If shuffle is enabled or not."""
        return self._shuffle

    @property
    def repeat(self):
        """Current repeat mode."""
        return self._repeat


class MetadataDummy(interface.Metadata):
    def artwork(self):
        """Return artwork for what is currently playing (or None)."""
        raise exceptions.NotSupportedError()

    def artwork_id(self):
        """Return a unique identifier for current artwork."""
        raise exceptions.NotSupportedError()

    def playing(self):
        """Return what is currently playing."""
        raise exceptions.NotSupportedError()

    @property
    def app(self):
        """Return information about running app."""
        raise exceptions.NotSupportedError()


class FeaturesDummy(interface.Features):
    def __init__(self, feature_map: Dict[FeatureName, FeatureState]):
        self.feature_map = feature_map

    def get_feature(self, feature: FeatureName) -> FeatureInfo:
        """Return current state of a feature."""
        if feature not in self.feature_map:
            state = FeatureState.Unsupported
        else:
            state = self.feature_map[feature]
        return FeatureInfo(state=state)


@pytest.fixture
def methods():
    yield interface.retrieve_commands(TestClass)


# COMMANDS


def test_get_commands(methods):
    assert len(methods) == 5
    assert "test_method" in methods
    assert "another_method" in methods
    assert "some_property" in methods
    assert "abbrev_help" in methods


def test_get_first_sentence_without_leading_period_in_pydoc(methods):
    assert "Help text" == methods["test_method"]
    assert "Some other help text" == methods["another_method"]
    assert "Property help" == methods["some_property"]


def test_try_to_be_smart_with_abbreviations(methods):
    assert "Type, e.g. a, b or c" == methods["abbrev_help"]
    assert "Type, e.g. a, b or c" == methods["abbrev_help_more_text"]


# PLAYING


def test_playing_media_type_and_playstate():
    out = str(PlayingDummy())
    assert convert.media_type_str(MediaType.Video) in out
    assert convert.device_state_str(DeviceState.Playing) in out


def test_playing_title_artist_album_genre():
    out = str(
        PlayingDummy(
            title="mytitle", artist="myartist", album="myalbum", genre="mygenre"
        )
    )
    assert "mytitle" in out
    assert "myartist" in out
    assert "myalbum" in out
    assert "mygenre" in out


def test_playing_only_position():
    assert "1234" in str(PlayingDummy(position=1234))


def test_playing_only_total_time():
    assert "5678" in str(PlayingDummy(total_time=5678))


def test_playing_both_position_and_total_time():
    out = str(PlayingDummy(position=1234, total_time=5678))
    assert "1234/5678" in out


def test_playing_shuffle_and_repeat():
    out = str(PlayingDummy())
    assert "Shuffle: Songs" in out
    assert "Repeat: Track" in out


def test_playing_generate_same_hash():
    playing = PlayingDummy(
        title="title", artist="artist", album="album", total_time=123
    )
    assert (
        "538df531d1715629fdd87affd0c5957bcbf54cd89180778071e6535b7df4e22c"
        == playing.hash
    )

    playing2 = PlayingDummy(title="dummy", artist="test", album="none", total_time=321)

    assert (
        "80045c05d18382f33a5369fd5cdfc6ae42c3eb418125f638d7a31ab173b01ade"
        == playing2.hash
    )


# METADATA


def test_metadata_device_id():
    assert MetadataDummy("dummy").device_id == "dummy"


def test_metadata_rest_not_supported():
    metadata = MetadataDummy("dummy")

    with pytest.raises(exceptions.NotSupportedError):
        metadata.artwork()
    with pytest.raises(exceptions.NotSupportedError):
        metadata.artwork_id()
    with pytest.raises(exceptions.NotSupportedError):
        metadata.playing()


# DEVICE INFO


def test_fields_set():
    dev_info = interface.DeviceInfo(
        OperatingSystem.TvOS, "1.2.3", "19A123", DeviceModel.Gen4K, "aa:bb:cc:dd:ee:ff",
    )

    assert dev_info.operating_system == OperatingSystem.TvOS
    assert dev_info.version == "1.2.3"
    assert dev_info.build_number == "19A123"
    assert dev_info.model == DeviceModel.Gen4K
    assert dev_info.mac == "aa:bb:cc:dd:ee:ff"


def test_apple_tv_software_str():
    dev_info = interface.DeviceInfo(
        OperatingSystem.Legacy,
        "2.2.3",
        "13D333",
        DeviceModel.Gen3,
        "aa:bb:cc:dd:ee:ff",
    )

    assert str(dev_info) == "3 ATV SW 2.2.3 build 13D333"


def test_tvos_str():
    dev_info = interface.DeviceInfo(
        OperatingSystem.TvOS, "1.2.3", "19A123", DeviceModel.Gen4K, "aa:bb:cc:dd:ee:ff",
    )

    assert str(dev_info) == "4K tvOS 1.2.3 build 19A123"


def test_unknown_str():
    dev_info = interface.DeviceInfo(
        OperatingSystem.Unknown, None, None, DeviceModel.Unknown, None
    )

    assert str(dev_info) == "Unknown Model Unknown OS"


# FEATURES


def test_all_unsupported_features():
    features = FeaturesDummy({FeatureName.Play: FeatureState.Unsupported})
    assert not features.all_features()


def test_all_include_unsupported_features():
    features = FeaturesDummy({FeatureName.Play: FeatureState.Unsupported})
    all_features = features.all_features(include_unsupported=True)

    assert set(all_features.keys()) == set(FeatureName)
    assert set([ft.state for ft in all_features.values()]) == set(
        [FeatureState.Unsupported]
    )


def test_features_in_state():
    features = FeaturesDummy(
        {
            FeatureName.Play: FeatureState.Unsupported,
            FeatureName.Pause: FeatureState.Available,
        }
    )

    assert not features.in_state(FeatureState.Unknown, FeatureName.Play)
    assert not features.in_state([FeatureState.Unknown], FeatureName.Play)
    assert features.in_state(FeatureState.Unsupported, FeatureName.Play)
    assert features.in_state([FeatureState.Unsupported], FeatureName.Play)

    assert not features.in_state(
        [FeatureState.Unsupported], FeatureName.Play, FeatureName.Pause
    )

    assert features.in_state(
        [FeatureState.Unsupported, FeatureState.Available],
        FeatureName.Play,
        FeatureName.Pause,
    )


# APP


def test_app_properties():
    app = interface.App("name", "id")
    assert app.name == "name"
    assert app.identifier == "id"


def test_app_str():
    app = interface.App("name", "id")
    assert "App: name (id)" == str(app)



"""Simple web server for serving a file to stream via AirPlay."""
import pathlib
import logging

from aiohttp import web
from aiohttp.web import middleware

from pyatv.support.net import unused_port

_LOGGER = logging.getLogger(__name__)


class StaticFileWebServer:
    """Web server serving only a single file."""

    def __init__(self, file_to_serve, address, port=None):
        """Initialize a new StaticFileWebServer."""
        self.path = pathlib.Path(file_to_serve)
        self.app = web.Application(middlewares=[self._middleware])
        self.app.router.add_static("/", self.path.parent, show_index=False)
        self.runner = web.AppRunner(self.app)
        self.site = None
        self._address = address  # Local address to bind to
        self._port = port

    async def start(self):
        """Start the web server."""
        if not self._port:
            self._port = unused_port()

        _LOGGER.debug("Starting AirPlay file server on port %d", self._port)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, str(self._address), self._port)
        await self.site.start()

    async def close(self):
        """Stop the web server and free resources."""
        _LOGGER.debug("Closing local AirPlay web server")
        await self.runner.cleanup()

    @property
    def file_address(self):
        """Address to the file being served."""
        return f"http://{self._address}:{self._port}/{self.path.name}"

    # This middleware makes sure only the specified file is accessible. This is needed
    # since aiohttp only supports serving an entire directory.
    @middleware
    async def _middleware(self, request, handler):
        if request.rel_url.path == f"/{self.path.name}":
            return await handler(request)
        return web.Response(body="Permission denied", status=401)



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    FieldDescriptor as google___protobuf___descriptor___FieldDescriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Optional as typing___Optional,
    Union as typing___Union,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


builtin___bool = bool
builtin___bytes = bytes
builtin___float = float
builtin___int = int
if sys.version_info < (3,):
    builtin___buffer = buffer
    builtin___unicode = unicode


class CryptoPairingMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    pairingData = ... # type: builtin___bytes
    status = ... # type: builtin___int
    isRetrying = ... # type: builtin___bool
    isUsingSystemPairing = ... # type: builtin___bool
    state = ... # type: builtin___int

    def __init__(self,
        *,
        pairingData : typing___Optional[builtin___bytes] = None,
        status : typing___Optional[builtin___int] = None,
        isRetrying : typing___Optional[builtin___bool] = None,
        isUsingSystemPairing : typing___Optional[builtin___bool] = None,
        state : typing___Optional[builtin___int] = None,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> CryptoPairingMessage: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> CryptoPairingMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"isRetrying",b"isRetrying",u"isUsingSystemPairing",b"isUsingSystemPairing",u"pairingData",b"pairingData",u"state",b"state",u"status",b"status"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"isRetrying",b"isRetrying",u"isUsingSystemPairing",b"isUsingSystemPairing",u"pairingData",b"pairingData",u"state",b"state",u"status",b"status"]) -> None: ...
global___CryptoPairingMessage = CryptoPairingMessage

cryptoPairingMessage = ... # type: google___protobuf___descriptor___FieldDescriptor



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    EnumDescriptor as google___protobuf___descriptor___EnumDescriptor,
    FieldDescriptor as google___protobuf___descriptor___FieldDescriptor,
)

from google.protobuf.internal.containers import (
    RepeatedScalarFieldContainer as google___protobuf___internal___containers___RepeatedScalarFieldContainer,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from pyatv.mrp.protobuf.PlayerPath_pb2 import (
    PlayerPath as pyatv___mrp___protobuf___PlayerPath_pb2___PlayerPath,
)

from typing import (
    Iterable as typing___Iterable,
    List as typing___List,
    Optional as typing___Optional,
    Text as typing___Text,
    Tuple as typing___Tuple,
    Union as typing___Union,
    cast as typing___cast,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


builtin___bool = bool
builtin___bytes = bytes
builtin___float = float
builtin___int = int
builtin___str = str
if sys.version_info < (3,):
    builtin___buffer = buffer
    builtin___unicode = unicode


class SendError(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Enum(builtin___int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: builtin___int) -> builtin___str: ...
        @classmethod
        def Value(cls, name: builtin___str) -> 'SendError.Enum': ...
        @classmethod
        def keys(cls) -> typing___List[builtin___str]: ...
        @classmethod
        def values(cls) -> typing___List['SendError.Enum']: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[builtin___str, 'SendError.Enum']]: ...
        NoError = typing___cast('SendError.Enum', 0)
        ApplicationNotFound = typing___cast('SendError.Enum', 1)
        ConnectionFailed = typing___cast('SendError.Enum', 2)
        Ignored = typing___cast('SendError.Enum', 3)
        CouldNotLaunchApplication = typing___cast('SendError.Enum', 4)
        TimedOut = typing___cast('SendError.Enum', 5)
        OriginDoesNotExist = typing___cast('SendError.Enum', 6)
        InvalidOptions = typing___cast('SendError.Enum', 7)
        NoCommandHandlers = typing___cast('SendError.Enum', 8)
        ApplicationNotInstalled = typing___cast('SendError.Enum', 9)
    NoError = typing___cast('SendError.Enum', 0)
    ApplicationNotFound = typing___cast('SendError.Enum', 1)
    ConnectionFailed = typing___cast('SendError.Enum', 2)
    Ignored = typing___cast('SendError.Enum', 3)
    CouldNotLaunchApplication = typing___cast('SendError.Enum', 4)
    TimedOut = typing___cast('SendError.Enum', 5)
    OriginDoesNotExist = typing___cast('SendError.Enum', 6)
    InvalidOptions = typing___cast('SendError.Enum', 7)
    NoCommandHandlers = typing___cast('SendError.Enum', 8)
    ApplicationNotInstalled = typing___cast('SendError.Enum', 9)
    global___Enum = Enum


    def __init__(self,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> SendError: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> SendError: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
global___SendError = SendError

class HandlerReturnStatus(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Enum(builtin___int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: builtin___int) -> builtin___str: ...
        @classmethod
        def Value(cls, name: builtin___str) -> 'HandlerReturnStatus.Enum': ...
        @classmethod
        def keys(cls) -> typing___List[builtin___str]: ...
        @classmethod
        def values(cls) -> typing___List['HandlerReturnStatus.Enum']: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[builtin___str, 'HandlerReturnStatus.Enum']]: ...
        Success = typing___cast('HandlerReturnStatus.Enum', 0)
        NoSuchContent = typing___cast('HandlerReturnStatus.Enum', 1)
        CommandFailed = typing___cast('HandlerReturnStatus.Enum', 2)
        NoActionableNowPlayingItem = typing___cast('HandlerReturnStatus.Enum', 10)
        DeviceNotFound = typing___cast('HandlerReturnStatus.Enum', 20)
        UIKitLegacy = typing___cast('HandlerReturnStatus.Enum', 3)
        SkipAdProhibited = typing___cast('HandlerReturnStatus.Enum', 100)
        QueueIsUserCurated = typing___cast('HandlerReturnStatus.Enum', 101)
        UserModifiedQueueDisabled = typing___cast('HandlerReturnStatus.Enum', 102)
        UserQueueModificationNotSupportedForCurrentItem = typing___cast('HandlerReturnStatus.Enum', 103)
        SubscriptionRequiredForSharedQueue = typing___cast('HandlerReturnStatus.Enum', 104)
    Success = typing___cast('HandlerReturnStatus.Enum', 0)
    NoSuchContent = typing___cast('HandlerReturnStatus.Enum', 1)
    CommandFailed = typing___cast('HandlerReturnStatus.Enum', 2)
    NoActionableNowPlayingItem = typing___cast('HandlerReturnStatus.Enum', 10)
    DeviceNotFound = typing___cast('HandlerReturnStatus.Enum', 20)
    UIKitLegacy = typing___cast('HandlerReturnStatus.Enum', 3)
    SkipAdProhibited = typing___cast('HandlerReturnStatus.Enum', 100)
    QueueIsUserCurated = typing___cast('HandlerReturnStatus.Enum', 101)
    UserModifiedQueueDisabled = typing___cast('HandlerReturnStatus.Enum', 102)
    UserQueueModificationNotSupportedForCurrentItem = typing___cast('HandlerReturnStatus.Enum', 103)
    SubscriptionRequiredForSharedQueue = typing___cast('HandlerReturnStatus.Enum', 104)
    global___Enum = Enum


    def __init__(self,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> HandlerReturnStatus: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> HandlerReturnStatus: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
global___HandlerReturnStatus = HandlerReturnStatus

class SendCommandResultMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    sendError = ... # type: global___SendError.Enum
    handlerReturnStatus = ... # type: global___HandlerReturnStatus.Enum
    handlerReturnStatusDatas = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[builtin___bytes]
    commandID = ... # type: typing___Text

    @property
    def playerPath(self) -> pyatv___mrp___protobuf___PlayerPath_pb2___PlayerPath: ...

    def __init__(self,
        *,
        sendError : typing___Optional[global___SendError.Enum] = None,
        handlerReturnStatus : typing___Optional[global___HandlerReturnStatus.Enum] = None,
        handlerReturnStatusDatas : typing___Optional[typing___Iterable[builtin___bytes]] = None,
        commandID : typing___Optional[typing___Text] = None,
        playerPath : typing___Optional[pyatv___mrp___protobuf___PlayerPath_pb2___PlayerPath] = None,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> SendCommandResultMessage: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> SendCommandResultMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"commandID",b"commandID",u"handlerReturnStatus",b"handlerReturnStatus",u"playerPath",b"playerPath",u"sendError",b"sendError"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"commandID",b"commandID",u"handlerReturnStatus",b"handlerReturnStatus",u"handlerReturnStatusDatas",b"handlerReturnStatusDatas",u"playerPath",b"playerPath",u"sendError",b"sendError"]) -> None: ...
global___SendCommandResultMessage = SendCommandResultMessage

sendCommandResultMessage = ... # type: google___protobuf___descriptor___FieldDescriptor



"""Methods used to make GET/POST requests."""

import socket
import logging
from ipaddress import IPv4Interface, IPv4Address
from typing import Optional

import netifaces
from aiohttp import ClientSession

from pyatv.support import log_binary

_LOGGER = logging.getLogger(__name__)


# This timeout is rather long and that is for a reason. If a device is sleeping, it
# automatically wakes up when a service is requested from it. Up to 20 seconds or so
# have been seen. So to deal with that, keep this high.
DEFAULT_TIMEOUT = 25.0  # Seconds


def is_custom_session(session):
    """Return if a ClientSession was created by pyatv."""
    return hasattr(session, "_pyatv")


async def create_session(loop):
    """Create aiohttp ClientSession manged by pyatv."""
    session = ClientSession(loop=loop)
    setattr(session, "_pyatv", True)
    return session


def unused_port() -> int:
    """Return a port that is unused on the current host."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        return s.getsockname()[1]


def get_local_address_reaching(dest_ip: IPv4Address) -> Optional[IPv4Address]:
    """Get address of a local interface within same subnet as provided address."""
    for iface in netifaces.interfaces():
        for addr in netifaces.ifaddresses(iface).get(netifaces.AF_INET, []):
            iface = IPv4Interface(addr["addr"] + "/" + addr["netmask"])
            if dest_ip in iface.network:
                return iface.ip
    return None


class HttpSession:
    """This class simplifies GET/POST requests."""

    def __init__(self, client_session, base_url):
        """Initialize a new HttpSession."""
        self._session = client_session  # aiohttp session
        self.base_url = base_url

    async def get_data(self, path, headers=None, timeout=None):
        """Perform a GET request."""
        url = self.base_url + path
        _LOGGER.debug("GET URL: %s", url)
        resp = None
        try:
            resp = await self._session.get(
                url,
                headers=headers,
                timeout=DEFAULT_TIMEOUT if timeout is None else timeout,
            )
            _LOGGER.debug(
                "Response: status=%d, headers=[%s]",
                resp.status,
                ", ".join([f"{key}={value}" for key, value in resp.headers.items()]),
            )
            if resp.content_length is not None:
                resp_data = await resp.read()
                log_binary(_LOGGER, "<< GET", Data=resp_data)
            else:
                resp_data = None
            return resp_data, resp.status
        except Exception as ex:
            if resp is not None:
                resp.close()
            raise ex
        finally:
            if resp is not None:
                await resp.release()

    async def post_data(self, path, data=None, headers=None, timeout=None):
        """Perform a POST request."""
        url = self.base_url + path
        _LOGGER.debug("POST URL: %s", url)
        log_binary(_LOGGER, ">> POST", Data=data)

        resp = None
        try:
            resp = await self._session.post(
                url,
                headers=headers,
                data=data,
                timeout=DEFAULT_TIMEOUT if timeout is None else timeout,
            )
            _LOGGER.debug(
                "Response: status=%d, headers=[%s]",
                resp.status,
                ", ".join([f"{key}={value}" for key, value in resp.headers.items()]),
            )
            if resp.content_length is not None:
                resp_data = await resp.read()
            else:
                resp_data = None
                log_binary(_LOGGER, "<< POST", Data=resp_data)
            return resp_data, resp.status
        except Exception as ex:
            if resp is not None:
                resp.close()
            raise ex
        finally:
            if resp is not None:
                await resp.release()



"""Module used to knock on TCP ports.

This module will open a TCP-connection to one or more ports on a given host and
immediately close it. The use case for this is to wake a device sleeping via a Bonjour
sleep proxy. Such a device will automatically be woken up in case any of its services
are accessed, something this module will try to emulate.
"""

import math
import socket
import select
import asyncio
import logging

from typing import List
from ipaddress import IPv4Address

_LOGGER = logging.getLogger(__name__)

SEND_INTERVAL = 2.0


# Performs the actual "knock". This can most certainly be done with asyncio code, but
# works for now.
def _synch_knock(address: IPv4Address, port: int):
    _LOGGER.debug("Knocking at port %d on %s", port, address)

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setblocking(False)

    socket_address = (str(address), port)
    s.connect_ex(socket_address)
    select.select([s], [s], [s], 0.1)
    s.close()


async def knock(
    address: IPv4Address, ports: List[int], loop: asyncio.AbstractEventLoop
):
    """Knock on a set of ports for a given host."""
    await asyncio.wait(
        [loop.run_in_executor(None, _synch_knock, address, port) for port in ports]
    )


async def knocker(
    address: IPv4Address,
    ports: List[int],
    loop: asyncio.AbstractEventLoop,
    timeout: int = 4,
) -> asyncio.Future:
    """Continuously knock on a set of ports.

    New port knocks are sent every two seconds, so a timeout of 4 seconds will result in
    two knocks.
    """
    no_of_sends = math.ceil(timeout / SEND_INTERVAL)

    async def _repeat():
        for _ in range(no_of_sends):
            await knock(address, ports, loop)
            await asyncio.sleep(SEND_INTERVAL)

    return asyncio.ensure_future(_repeat())



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    EnumDescriptor as google___protobuf___descriptor___EnumDescriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    List as typing___List,
    Tuple as typing___Tuple,
    Union as typing___Union,
    cast as typing___cast,
)


builtin___bytes = bytes
builtin___int = int
builtin___str = str
if sys.version_info < (3,):
    builtin___buffer = buffer
    builtin___unicode = unicode


class RepeatMode(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Enum(builtin___int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: builtin___int) -> builtin___str: ...
        @classmethod
        def Value(cls, name: builtin___str) -> 'RepeatMode.Enum': ...
        @classmethod
        def keys(cls) -> typing___List[builtin___str]: ...
        @classmethod
        def values(cls) -> typing___List['RepeatMode.Enum']: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[builtin___str, 'RepeatMode.Enum']]: ...
        Unknown = typing___cast('RepeatMode.Enum', 0)
        One = typing___cast('RepeatMode.Enum', 1)
        All = typing___cast('RepeatMode.Enum', 2)
    Unknown = typing___cast('RepeatMode.Enum', 0)
    One = typing___cast('RepeatMode.Enum', 1)
    All = typing___cast('RepeatMode.Enum', 2)
    global___Enum = Enum


    def __init__(self,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> RepeatMode: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> RepeatMode: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
global___RepeatMode = RepeatMode

class ShuffleMode(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Enum(builtin___int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: builtin___int) -> builtin___str: ...
        @classmethod
        def Value(cls, name: builtin___str) -> 'ShuffleMode.Enum': ...
        @classmethod
        def keys(cls) -> typing___List[builtin___str]: ...
        @classmethod
        def values(cls) -> typing___List['ShuffleMode.Enum']: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[builtin___str, 'ShuffleMode.Enum']]: ...
        Unknown = typing___cast('ShuffleMode.Enum', 0)
        Off = typing___cast('ShuffleMode.Enum', 1)
        Albums = typing___cast('ShuffleMode.Enum', 2)
        Songs = typing___cast('ShuffleMode.Enum', 3)
    Unknown = typing___cast('ShuffleMode.Enum', 0)
    Off = typing___cast('ShuffleMode.Enum', 1)
    Albums = typing___cast('ShuffleMode.Enum', 2)
    Songs = typing___cast('ShuffleMode.Enum', 3)
    global___Enum = Enum


    def __init__(self,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> ShuffleMode: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> ShuffleMode: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
global___ShuffleMode = ShuffleMode

class DeviceClass(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Enum(builtin___int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: builtin___int) -> builtin___str: ...
        @classmethod
        def Value(cls, name: builtin___str) -> 'DeviceClass.Enum': ...
        @classmethod
        def keys(cls) -> typing___List[builtin___str]: ...
        @classmethod
        def values(cls) -> typing___List['DeviceClass.Enum']: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[builtin___str, 'DeviceClass.Enum']]: ...
        Invalid = typing___cast('DeviceClass.Enum', 0)
        iPhone = typing___cast('DeviceClass.Enum', 1)
        iPod = typing___cast('DeviceClass.Enum', 2)
        iPad = typing___cast('DeviceClass.Enum', 3)
        AppleTV = typing___cast('DeviceClass.Enum', 4)
        iFPGA = typing___cast('DeviceClass.Enum', 5)
        Watch = typing___cast('DeviceClass.Enum', 6)
        Accessory = typing___cast('DeviceClass.Enum', 7)
        Bridge = typing___cast('DeviceClass.Enum', 8)
        Mac = typing___cast('DeviceClass.Enum', 9)
    Invalid = typing___cast('DeviceClass.Enum', 0)
    iPhone = typing___cast('DeviceClass.Enum', 1)
    iPod = typing___cast('DeviceClass.Enum', 2)
    iPad = typing___cast('DeviceClass.Enum', 3)
    AppleTV = typing___cast('DeviceClass.Enum', 4)
    iFPGA = typing___cast('DeviceClass.Enum', 5)
    Watch = typing___cast('DeviceClass.Enum', 6)
    Accessory = typing___cast('DeviceClass.Enum', 7)
    Bridge = typing___cast('DeviceClass.Enum', 8)
    Mac = typing___cast('DeviceClass.Enum', 9)
    global___Enum = Enum


    def __init__(self,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> DeviceClass: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> DeviceClass: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
global___DeviceClass = DeviceClass

class DeviceType(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Enum(builtin___int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: builtin___int) -> builtin___str: ...
        @classmethod
        def Value(cls, name: builtin___str) -> 'DeviceType.Enum': ...
        @classmethod
        def keys(cls) -> typing___List[builtin___str]: ...
        @classmethod
        def values(cls) -> typing___List['DeviceType.Enum']: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[builtin___str, 'DeviceType.Enum']]: ...
        Unknown = typing___cast('DeviceType.Enum', 0)
        AirPlay = typing___cast('DeviceType.Enum', 1)
        Bluetooth = typing___cast('DeviceType.Enum', 2)
        CarPlay = typing___cast('DeviceType.Enum', 3)
        BuiltIn = typing___cast('DeviceType.Enum', 4)
        Wired = typing___cast('DeviceType.Enum', 5)
    Unknown = typing___cast('DeviceType.Enum', 0)
    AirPlay = typing___cast('DeviceType.Enum', 1)
    Bluetooth = typing___cast('DeviceType.Enum', 2)
    CarPlay = typing___cast('DeviceType.Enum', 3)
    BuiltIn = typing___cast('DeviceType.Enum', 4)
    Wired = typing___cast('DeviceType.Enum', 5)
    global___Enum = Enum


    def __init__(self,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> DeviceType: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> DeviceType: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
global___DeviceType = DeviceType

class DeviceSubType(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Enum(builtin___int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: builtin___int) -> builtin___str: ...
        @classmethod
        def Value(cls, name: builtin___str) -> 'DeviceSubType.Enum': ...
        @classmethod
        def keys(cls) -> typing___List[builtin___str]: ...
        @classmethod
        def values(cls) -> typing___List['DeviceSubType.Enum']: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[builtin___str, 'DeviceSubType.Enum']]: ...
        Default = typing___cast('DeviceSubType.Enum', 0)
        Speaker = typing___cast('DeviceSubType.Enum', 1)
        Headphones = typing___cast('DeviceSubType.Enum', 2)
        Headset = typing___cast('DeviceSubType.Enum', 3)
        Receiver = typing___cast('DeviceSubType.Enum', 4)
        LineOut = typing___cast('DeviceSubType.Enum', 5)
        USB = typing___cast('DeviceSubType.Enum', 6)
        DisplayPort = typing___cast('DeviceSubType.Enum', 7)
        HDMI = typing___cast('DeviceSubType.Enum', 8)
        LowEnergy = typing___cast('DeviceSubType.Enum', 9)
        SPDIF = typing___cast('DeviceSubType.Enum', 10)
        TV = typing___cast('DeviceSubType.Enum', 11)
        HomePod = typing___cast('DeviceSubType.Enum', 12)
        AppleTV = typing___cast('DeviceSubType.Enum', 13)
    Default = typing___cast('DeviceSubType.Enum', 0)
    Speaker = typing___cast('DeviceSubType.Enum', 1)
    Headphones = typing___cast('DeviceSubType.Enum', 2)
    Headset = typing___cast('DeviceSubType.Enum', 3)
    Receiver = typing___cast('DeviceSubType.Enum', 4)
    LineOut = typing___cast('DeviceSubType.Enum', 5)
    USB = typing___cast('DeviceSubType.Enum', 6)
    DisplayPort = typing___cast('DeviceSubType.Enum', 7)
    HDMI = typing___cast('DeviceSubType.Enum', 8)
    LowEnergy = typing___cast('DeviceSubType.Enum', 9)
    SPDIF = typing___cast('DeviceSubType.Enum', 10)
    TV = typing___cast('DeviceSubType.Enum', 11)
    HomePod = typing___cast('DeviceSubType.Enum', 12)
    AppleTV = typing___cast('DeviceSubType.Enum', 13)
    global___Enum = Enum


    def __init__(self,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> DeviceSubType: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> DeviceSubType: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
global___DeviceSubType = DeviceSubType
